"""
imephu is a package for generating astronomical finder charts.

The finder charts are primarily intended for the Southern African Large Telescope
(SALT).
"""

from importlib.metadata import version

__version__ = version("imephu")
